import math

PHI = (math.sqrt(5.0)+1) / 2

class eval_context(object):
    def __init__(self):
        self.frame_stack = []
    def get_var_value(self, name):
        if (name == 'PI'):
            return math.pi
        if (name == 'PHI'):
            global PHI
            return PHI
        curr_frame = self.frame_stack[-1]
        return curr_frame[name]
    def push_frame(self, frame):
        self.frame_stack.append( frame )
    def pop_frame(self):
        self.frame_stack.pop()
    def get_stack_depth(self):
        return len(self.frame_stack)

#------------------------------------------------------------------------------#

class geometry_context(object):
    class Polygon(object):
        def __init__(self, fill_color, stroke_color, thickness, start_pt):
            self.fill_color = fill_color
            self.stroke_color = stroke_color
            self.pts = [ start_pt ]
            self.thickness = thickness
        def insert( self, pt ):
            self.pts.append( pt )
        def to_svg(self):
            pts_string = ' '.join( [ "%.6f,%.6f" % (x,y) for (x,y) in self.pts] )
            fill = "none"
            stroke = "none"
            if self.fill_color:
                fill = self.fill_color if self.fill_color[0] != '"' else self.fill_color[1:-1]
            if self.stroke_color:
                stroke = self.stroke_color if self.stroke_color[0] != '"' else self.stroke_color[1:-1]
            return "<polyline points='%s' style='fill:%s;stroke:%s;stroke-width:%s;shape-rendering:crispEdges;' />\n" % (pts_string, fill, stroke, self.thickness)

    def __init__(self, tile_dict, start_pt, max_depth, scale):
        self.tile_dict = tile_dict
        self.turtle_state_stack = [{ 'loc':start_pt, 'theta':0 }]
        self.polygons = []
        self.curr_polygon_stack = []
        self.closed_polys = []
        self.eval_ctxt = eval_context()
        self.max_depth = max_depth
        self.scale = scale
        self.flip_factor = 1.0

    def flip(self):
        self.flip_factor *= -1.0

    def get_curr_loc(self):
        turtle_state = self.turtle_state_stack[-1]
        return turtle_state['loc']

    def get_curr_dir(self):
        turtle_state = self.turtle_state_stack[-1]
        return turtle_state['theta']

    def set_curr_loc(self, loc):
        turtle_state = self.turtle_state_stack[-1]
        turtle_state['loc'] = loc

    def set_curr_dir(self, theta):
        turtle_state = self.turtle_state_stack[-1]
        turtle_state['theta'] = theta

    def push_turtle_state(self, loc, theta):
        self.turtle_state_stack.append({ 'loc':loc, 'theta':theta })

    def pop_turtle_state( self ):
        self.turtle_state_stack.pop()

    def get_curr_depth(self):
        return self.eval_ctxt.get_stack_depth()

    def push_frame(self, parameters, values):
        frame = { var : val for (var, val) in zip(parameters, values)}
        self.eval_ctxt.push_frame( frame )

    def pop_frame(self):
        self.eval_ctxt.pop_frame()

    def turn(self, angle):
        self.set_curr_dir( self.get_curr_dir() + self.flip_factor*angle )

    def forward(self, dist):
        x = self.get_curr_loc()[0]
        y = self.get_curr_loc()[1]
        self.set_curr_loc( ( x + self.scale * dist * math.cos( self.get_curr_dir() ), y + self.scale * dist * math.sin( self.get_curr_dir())) )
        self.add_pt_to_curr_poly( self.get_curr_loc() )

    def open_poly(self, fill_color, stroke_color, thickness):
        self.curr_polygon_stack.append(
            self.Polygon( fill_color, stroke_color, thickness, self.get_curr_loc())
        )

    def close_poly(self):
        if len(self.curr_polygon_stack) == 0:
            raise Exception("Attempted to close nonexistent polygon")
        poly = self.curr_polygon_stack.pop()
        # The point of this business with "closed_polys" is to add nested polygon objects to
        # the output list such that the most deeply nested polys will be on top ... this way
        # we ensure that if you stroke on top of a fill by nesting a stroke in a fill block
        # the stroke won't be occluded by the fill...
        self.closed_polys = [poly] + self.closed_polys
        if (len(self.curr_polygon_stack) == 0):
            self.polygons += self.closed_polys
            self.closed_polys = []

    def add_pt_to_curr_poly(self, pt):
        for poly in self.curr_polygon_stack:
            poly.insert(pt)

    def to_svg(self, use_html, wd, hgt):
        output = ""
        if use_html:
            output += "<!DOCTYPE html>\n<html>\n<body>\n"
        output += "<svg width='%d' height='%d'>\n" % (wd, hgt)
        for poly in self.polygons:
            output += poly.to_svg()
        output += "</svg>\n"
        if use_html:
            output += "</body></html>"
        return output

#------------------------------------------------------------------------------#

class expr_num(object):
    def __init__(self, num):
        self.num = float(num)
    def eval(self, ctxt):
        return self.num

class expr_negate(object):
    def __init__(self, child_expr):
        self.expr = child_expr
    def eval(self,ctxt):
        return  - self.expr.eval(ctxt)

class expr_var(object):
    def __init__(self, name):
        self.var = name
    def eval(self, ctxt):
        return ctxt.get_var_value( self.var )

class expr(object):
    def __init__(self, children, ops):
        self.children = children
        self.ops = ops
    def eval(self, ctxt):
        if ( len(self.children) == 1 ):
            return self.children[0].eval(ctxt)
        else:
            return self.do_ops(ctxt)
    def do_ops(self, ctxt):
        output = self.children[0].eval(ctxt)
        n = len(self.children)
        for i in range(1,n):
            op = self.ops[i-1]
            val = self.children[i].eval(ctxt)
            if (op == "+"):
                output += val
            elif (op == "-"):
                output -= val
            elif (op == "*"):
                output *= val
            elif (op == "/"):
                output /= val
        return output

#------------------------------------------------------------------------------#

class turn_stmt(object):
    def __init__(self, expr):
        self.angle_expr = expr
    def execute(self, geo_ctxt):
        theta = self.angle_expr.eval( geo_ctxt.eval_ctxt )
        geo_ctxt.turn( theta )

class pass_stmt(object):
    def __init__(self):
        pass
    def execute(self, geo_ctxt):
        pass

class forward_stmt(object):
    def __init__(self, expr):
        self.dist_expr = expr
    def execute(self, geo_ctxt):
        dist = self.dist_expr.eval( geo_ctxt.eval_ctxt )
        geo_ctxt.forward( dist )

class subroutine_stmt(object):
    def __init__(self, name, args):
        self.name = name
        self.args = args
    def execute(self, geo_ctxt):
        evaluated_args = list(map(lambda expr: expr.eval(geo_ctxt.eval_ctxt), self.args))
        tile = geo_ctxt.tile_dict[ self.name ]
        if (not tile):
            raise Exception("Unknown tile: " + self.name)
        tile.execute(geo_ctxt, evaluated_args)

class branch_stmt(object):
    def __init__(self, body):
        self.body = body
    def execute(self, geo_ctxt):
        loc = geo_ctxt.get_curr_loc()
        theta = geo_ctxt.get_curr_dir()
        geo_ctxt.push_turtle_state(loc, theta)
        for stmt in self.body:
            stmt.execute( geo_ctxt )
        geo_ctxt.pop_turtle_state()

class flip_stmt(object):
    def __init__(self, body):
        self.body = body
    def execute(self, geo_ctxt):
        geo_ctxt.flip()
        for stmt in self.body:
            stmt.execute( geo_ctxt )
        geo_ctxt.flip()

class fill_stmt(object):
    def __init__(self, color, body):
        self.fill_color = color
        self.body = body
    def execute(self, geo_ctxt):
        geo_ctxt.open_poly( self.fill_color, None, 0 )
        for stmt in self.body:
            stmt.execute( geo_ctxt )
        geo_ctxt.close_poly()

class stroke_stmt(object):
    def __init__(self, color, thickness, body):
        self.color = color
        self.thickness_expr = thickness
        self.body = body
    def execute(self, geo_ctxt):
        thickness = self.thickness_expr.eval(geo_ctxt.eval_ctxt)
        geo_ctxt.open_poly( None, self.color, thickness )
        for stmt in self.body:
            stmt.execute( geo_ctxt )
        geo_ctxt.close_poly()

class poly_stmt(object):
    def __init__(self, fill_color, stroke_color, thickness, body):
        self.fill_color = fill_color
        self.stroke_color = stroke_color
        self.thickness_expr = thickness
        self.body = body
    def execute(self, geo_ctxt):
        thickness = self.thickness_expr.eval(geo_ctxt.eval_ctxt)
        geo_ctxt.open_poly(self.fill_color, self.stroke_color, thickness )
        for stmt in self.body:
            stmt.execute( geo_ctxt )
        geo_ctxt.close_poly()

class substitution_def(object):
    def __init__(self, name, parameters, terminal_body, nonterminal_body):
        self.name = name
        self.parameters = parameters
        self.terminal_body = terminal_body
        self.nonterminal_body = nonterminal_body
    def execute(self, ctxt, values):
        ctxt.push_frame(self.parameters, values)
        if (ctxt.get_curr_depth() < ctxt.max_depth):
            for stmt in self.nonterminal_body:
                stmt.execute( ctxt )
        else:
            for stmt in self.terminal_body:
                stmt.execute( ctxt )
        ctxt.pop_frame();

class snurtle_script(object):
    def __init__(self, tile_defs):
        self.tile_defs = {tile.name: tile for tile in tile_defs}
    def get_tile_def(self, name):
        return self.tile_defs[name]
    def generate(self, tile_name, start_pt, levels, scale):
        start_tile = self.get_tile_def( tile_name )
        geo_ctxt = geometry_context(self.tile_defs, start_pt, levels, scale)
        if (start_tile):
            #TODO: make this work with lists of dimensions
            start_tile.execute(geo_ctxt, [1.0])
        return geo_ctxt
